/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef INTEGRALS_H
#define INTEGRALS_H

#include <cstddef>
#include <iostream>
#include <sstream>
#include <string>
#include <array>
#include <vector>
#include <unordered_set>
#include <unordered_map>
#include <memory>
#include <iterator>
#include <functional> // hash
#include <mutex>
#include <exception>

#include "defs.h"

namespace pyred {

/******************
 * Seed generator *
 ******************/

using compos_it_type = std::vector<std::vector<uint>>::const_iterator;
using compos_rit_type = std::vector<std::vector<uint>>::const_reverse_iterator;

class SeedIterator;

class Seeds {
public:
  typedef SeedIterator const_iterator;
  Seeds(const Seeds&) = default;
  Seeds(Seeds&&) = default;
  Seeds(uint, uint, uint, uint);
  SeedIterator begin() const;
  SeedIterator cbegin() const;
  SeedIterator end() const;
  SeedIterator cend() const;
private:
  std::vector<bool> sector;
  compos_rit_type ppows_begin, ppows_end;
  compos_it_type spows_begin, spows_end;
};

class SeedIterator {
public:
  typedef std::ptrdiff_t difference_type;
  typedef std::vector<pow_type> value_type;
  typedef const std::vector<pow_type>* pointer;
  typedef const std::vector<pow_type>& reference;
  typedef std::input_iterator_tag iterator_category;
  SeedIterator();
  SeedIterator(const SeedIterator&) = default;
  SeedIterator(SeedIterator&&) = default;
  SeedIterator(const std::vector<bool>& sector,
               const compos_rit_type&, const compos_rit_type&,
               const compos_it_type&, const compos_it_type&);
  SeedIterator& operator=(const SeedIterator&) = default;
  SeedIterator& operator++();
  SeedIterator operator++(int);
  std::vector<pow_type> operator*() const;
  std::unique_ptr<std::vector<pow_type>> operator->() const;
  friend bool operator==(const SeedIterator&, const SeedIterator&);
  friend void swap(SeedIterator&, SeedIterator&);
private:
  std::vector<bool> sector;
  compos_rit_type p_begin, p_it, p_end;
  compos_it_type s_it, s_end;
};

bool operator==(const SeedIterator&, const SeedIterator&);
bool operator!=(const SeedIterator&, const SeedIterator&);
void swap(SeedIterator&, SeedIterator&);

/******************
 * IntegralPyred class *
 ******************/

class IntegralPyredProperties {
public:
  uint topology;
  uint sector;
  uint lines;
  uint nums;
  uint dots;
  uint sps;
  uint ppows_weight;
  uint spows_weight;
};

class IntegralPyred {
public:
  IntegralPyred();
  IntegralPyred(uint, const std::vector<pow_type>&);
  IntegralPyred(uint, std::vector<pow_type>&&);
  IntegralPyred(std::istringstream&, uint);
  IntegralPyred(weight_type);
  IntegralPyred(const IntegralPyred&) = default;
  IntegralPyred(IntegralPyred&&) = default;
  uint topology;
  std::vector<pow_type> powers;
  weight_type to_weight() const;
  void reserve(std::size_t);
  std::size_t size() const;
  IntegralPyred zip(const std::vector<pow_type>&) const;
  IntegralPyred& operator=(const IntegralPyred&) = default;
  bool operator<(const IntegralPyred&) const;

  static void setup(uint, const std::vector<std::unordered_set<uint>>& = {},
                    uint = 1, const std::vector<uint>& = {5,4,17,13});
  static void setup(uint, const std::vector<std::string>&,
                    uint = 1, const std::vector<uint>& = {5,4,17,13});
  static uint np();
  static void use_cache(int=3);
  static void clear_cache(int=3);
  static uint sector_to_subweight(uint s) {
    return s_sector_weight[s].first;
  }
  static uint sector_from_subweight(uint w) {
    return s_sector_weight[w].second;
  }
  static IntegralPyredProperties properties(weight_type);
  static Seeds seeds(uint, uint, uint);
private:
  static std::pair<uint,uint> weight_to_dots(uint, uint, uint);
  static void weight_overflow(uint, uint, uint, uint, uint, uint);
  static uint s_np;
  static uint s_ordering;
  static std::vector<std::pair<uint,uint>> s_sector_weight;
  static uint minlines;
  // topology | sector | dots+sps | dots | ppows | spows || sum
  //     rest |   s_np |        5 |    4 |    17 |    13 ||  64
  static std::array<uint,6> s_weight_bits;
  static std::array<weight_type,6> s_weight_proj;
  static std::vector<std::unordered_set<uint>> s_trivialsectors;
  // cache_level: 0=(no cache), 1=(weight to integral only),
  //              2=(integral to weight only), 3=(both)
  static int cache_level;
  static std::unordered_map<IntegralPyred,weight_type> i2w_cache;
  static std::unordered_map<weight_type,IntegralPyred> w2i_cache;
  static std::mutex i2w_mtx;
  static std::mutex w2i_mtx;
  friend bool operator==(const IntegralPyred&, const IntegralPyred&);
  friend bool operator!=(const IntegralPyred&, const IntegralPyred&);
};


inline SeedIterator Seeds::begin() const {
  return SeedIterator(sector, ppows_begin, ppows_end, spows_begin, spows_end);
}

inline SeedIterator Seeds::end() const {
  return SeedIterator();
}

inline SeedIterator Seeds::cbegin() const {
  return begin();
}

inline SeedIterator Seeds::cend() const {
  return end();
}

inline void IntegralPyred::reserve(std::size_t sz) {
  powers.reserve(sz);
}

inline std::size_t IntegralPyred::size() const {
  return powers.size();
}

std::ostream& operator<<(std::ostream&, const IntegralPyred&);

}


namespace std {
  template<>
  struct hash<pyred::IntegralPyred> {
    size_t operator()(const pyred::IntegralPyred& igl) const {
      size_t h = igl.topology + 0x9e3779b9;
      for (const auto& pow: igl.powers) {
        h ^= pow + 0x9e3779b9 + (h << 6) + (h >> 2);
      }
      return h;
    }
  };
}

#endif
